#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import time
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional

from pychromecast.controllers.media import MEDIA_PLAYER_STATE_BUFFERING, MEDIA_PLAYER_STATE_IDLE, \
    MEDIA_PLAYER_STATE_PAUSED, \
    MEDIA_PLAYER_STATE_PLAYING, MEDIA_PLAYER_STATE_UNKNOWN, MediaStatus
from pychromecast.controllers.receiver import CastStatus
from pychromecast.socket_client import CONNECTION_STATUS_CONNECTED, CONNECTION_STATUS_CONNECTING, \
    CONNECTION_STATUS_DISCONNECTED, CONNECTION_STATUS_FAILED, CONNECTION_STATUS_FAILED_RESOLVE, CONNECTION_STATUS_LOST, \
    ConnectionStatus

from kast.utils.chrono import Seconds

DeviceName = str
VolumeLevel = float


class CastConnectionState(Enum):
    Connecting = CONNECTION_STATUS_CONNECTING
    Connected = CONNECTION_STATUS_CONNECTED
    Disconnected = CONNECTION_STATUS_DISCONNECTED
    Failed = CONNECTION_STATUS_FAILED
    FailedResolve = CONNECTION_STATUS_FAILED_RESOLVE
    Lost = CONNECTION_STATUS_LOST


class CastPlayerState(Enum):
    Playing = MEDIA_PLAYER_STATE_PLAYING
    Buffering = MEDIA_PLAYER_STATE_BUFFERING
    Paused = MEDIA_PLAYER_STATE_PAUSED
    Idle = MEDIA_PLAYER_STATE_IDLE
    Unknown = MEDIA_PLAYER_STATE_UNKNOWN


@dataclass
class CastCapabilities:
    canPause: bool = False
    canSeek: bool = False
    canSetMute: bool = False
    canSetVolume: bool = False
    canSkipForward: bool = False
    canSkipBackward: bool = False
    canQueueNext: bool = False
    canQueuePrevious: bool = False


@dataclass
class CastMediaState:
    volumeMuted: bool = False
    volumeLevel: VolumeLevel = 1.0
    playerState: CastPlayerState = CastPlayerState.Unknown
    duration: Seconds = Seconds()
    currentTime: Seconds = Seconds()
    title: str = ''
    displayName: str = ''
    iconUrl: str = ''
    imageUrl: str = ''
    contentUrl: Optional[str] = None
    lastPositionUpdateTimestamp: Optional[Seconds] = None

    @property
    def adjustedPosition(self) -> Seconds:
        return self.currentTime + (Seconds(time.time()) - self.lastPositionUpdateTimestamp) \
            if self.lastPositionUpdateTimestamp is not None else self.currentTime

    def isPlaying(self) -> bool:
        return self.playerState in [CastPlayerState.Playing, CastPlayerState.Buffering]

    def isPaused(self) -> bool:
        return self.playerState == CastPlayerState.Paused

    def isStopped(self) -> bool:
        return not self.isPlaying() and not self.isPaused()


@dataclass
class CastState:
    connection: CastConnectionState = CastConnectionState.Disconnected
    deviceName: Optional[DeviceName] = None
    capabilities: CastCapabilities = CastCapabilities()
    mediaState: CastMediaState = CastMediaState()

    def isConnected(self) -> bool:
        return self.connection in [
            CastConnectionState.Connected,
            CastConnectionState.Connecting
        ]

    def isConnectedOrRecoverable(self) -> bool:
        return self.isConnected() \
           or self.connection == CastConnectionState.Lost

    def update(
        self,
        mediaStatus: Optional[MediaStatus] = None,
        castStatus: Optional[CastStatus] = None,
        connectionStatus: Optional[ConnectionStatus] = None
    ) -> None:
        def validateValue(value: Any, defaultValue: Any) -> Any:
            return defaultValue if value is None else value

        if connectionStatus:
            self.connection = CastConnectionState(validateValue(connectionStatus.status, self.connection.value))

        if castStatus:
            self.mediaState.volumeMuted = validateValue(castStatus.volume_muted, self.mediaState.volumeMuted)
            self.mediaState.volumeLevel = validateValue(castStatus.volume_level, self.mediaState.volumeLevel)
            self.mediaState.displayName = validateValue(castStatus.display_name, self.mediaState.displayName)
            self.mediaState.iconUrl = validateValue(castStatus.icon_url, self.mediaState.iconUrl)

        if mediaStatus:
            self.capabilities.canPause = validateValue(mediaStatus.supports_pause, self.capabilities.canPause)
            self.capabilities.canSeek = validateValue(mediaStatus.supports_seek, self.capabilities.canSeek)
            self.capabilities.canSetMute = validateValue(mediaStatus.supports_stream_mute, self.capabilities.canSetMute)
            self.capabilities.canSetVolume = validateValue(mediaStatus.supports_stream_volume, self.capabilities.canSetVolume)
            self.capabilities.canSkipForward = validateValue(mediaStatus.supports_skip_forward, self.capabilities.canSkipForward)
            self.capabilities.canSkipBackward = validateValue(mediaStatus.supports_skip_backward, self.capabilities.canSkipBackward)
            self.capabilities.canQueueNext = validateValue(mediaStatus.supports_queue_next, self.capabilities.canQueueNext)
            self.capabilities.canQueuePrevious = validateValue(mediaStatus.supports_queue_prev, self.capabilities.canQueuePrevious)

            self.mediaState.volumeMuted = validateValue(mediaStatus.volume_muted, self.mediaState.volumeMuted)
            self.mediaState.volumeLevel = validateValue(mediaStatus.volume_level, self.mediaState.volumeLevel)
            self.mediaState.playerState = CastPlayerState(validateValue(mediaStatus.player_state, self.mediaState.playerState.value))
            self.mediaState.duration = Seconds(validateValue(mediaStatus.duration, self.mediaState.duration.value))
            self.mediaState.currentTime = Seconds(validateValue(mediaStatus.adjusted_current_time, self.mediaState.currentTime.value))
            self.mediaState.title = validateValue(mediaStatus.title, self.mediaState.title)
            self.mediaState.imageUrl = validateValue(mediaStatus.images[0].url if mediaStatus.images else '', self.mediaState.imageUrl)
            self.mediaState.contentUrl = validateValue(mediaStatus.content_id, self.mediaState.contentUrl)

            self.mediaState.lastPositionUpdateTimestamp = Seconds(time.time())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import logging
import logging.handlers
import platform
import signal
import sys
import threading
import types
from typing import Callable

import setproctitle

from kast.Services import Services
from kast.interface.qt5.KastQtApp import KastQtApp
from kast.utils.Loggable import Loggable


class KastApp(Loggable):

    def __init__(self, debug: bool = False) -> None:
        self._debug = debug
        self._services = Services()
        self._desktopApp = KastQtApp(services=self._services)

    def run(self) -> int:
        setproctitle.setproctitle(self._services.settings.appName)

        self._initLogger()

        signal.signal(signal.SIGTERM, self._onSignal)
        signal.signal(signal.SIGINT, self._onSignal)

        try:
            self.log.info(f"Platform: {platform.platform()}")
            self.log.info(f"Python: {sys.version}")
            self.log.info(f"Application: {self._services.settings.appName} ({self._services.settings.appVersion})")
            self.log.info(f"Persistent storage path: '{self._services.settings.persistentStoragePath}'")
            self.log.info(f"Temporary storage path: '{self._services.settings.temporaryStoragePath}'")

            self._services.mediaServer.mediaContent.thumbnailFile = self._services.settings.appIconPath
            self._services.mediaServer.start()
            self._runAsThread(self._services.osMediaOverlay.run)
            return self._desktopApp.run()

        except Exception as ex:
            self.log.exception(ex)
            return 1

        finally:
            self._onExit()

    def _onSignal(self, signum: int, frame: types.FrameType) -> None:
        self.log.info(f"Caught signal: {signal.Signals(signum).name}")
        self._desktopApp.exit(1)

    def _onExit(self) -> None:
        self._services.castController.quit()
        self._services.mediaServer.stop()

    def _runAsThread(self, callback: Callable[[], None]) -> None:
        thr = threading.Thread(target=callback)
        thr.daemon = True
        thr.start()

    def _initLogger(self) -> None:
        logFormatter = logging.Formatter(
            fmt='[%(asctime)s][%(threadName)s][%(levelname)-8.8s][%(name)s]: %(message)s',
            datefmt='%d/%m/%Y %I:%M:%S %p'
        )
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.INFO)

        fileHandler = logging.handlers.RotatingFileHandler(
            filename=self._services.settings.logFilePath,
            maxBytes=(5 * 1_000_000),
            backupCount=3
        )
        fileHandler.setFormatter(logFormatter)
        rootLogger.addHandler(fileHandler)

        if self._debug:
            consoleHandler = logging.StreamHandler()
            consoleHandler.setFormatter(logFormatter)
            rootLogger.addHandler(consoleHandler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import argparse
import sys
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Optional, Tuple, Any, List

DIR_ROOT = Path(__file__).resolve().parent


def makeReqFilePath(suffix: Optional[str] = None) -> Path:
    suffix = f'_{suffix}' if suffix else ''
    return DIR_ROOT / f'requirements{suffix}.txt'


@dataclass
class Package:
    name: str
    version: Optional[str] = None
    osName: Optional[str] = None


Packages = OrderedDict[str, Package]


def pkg(*args, **kwargs) -> Tuple[str, Package]:
    package = Package(*args, **kwargs)
    return package.name, package


@dataclass
class Requirements:
    filePath: Path
    packages: Packages


requirementsPkg = Requirements(
    filePath=makeReqFilePath(),
    packages=OrderedDict([
        pkg('tendo', '0.2.15'),
        pkg('psutil', '5.8.0'),
        pkg('av', '8.0.3'),
        pkg('pysubs2', '1.2.0'),
        pkg('pycaption', '1.0.5'),
        pkg('bottle', '0.12.19'),
        pkg('zeroconf', '0.30.0'),
        pkg('pychromecast', '9.2.0'),
        pkg('dataclasses-json', '0.5.4'),
        pkg('setproctitle', '1.2.2'),
        pkg('pyqt5', '5.15.2'),

        pkg('mpris_server', '0.3.4', 'posix'),

        pkg('winrt', '1.0.21033.1', 'nt'),
    ]))

requirementsDev = Requirements(
    filePath=makeReqFilePath('dev'),
    packages=OrderedDict([
        pkg('setuptools'),
        pkg('setup-qt', '1.0.0'),
    ]))

requirementsDist = Requirements(
    filePath=makeReqFilePath('dist'),
    packages=OrderedDict([
        pkg('twine'),
        pkg('PyInstaller'),

        pkg('Pillow', osName='nt'),
    ]))


class EnumBase(Enum):

    def __str__(self):
        return self.name.lower()

    def __call__(self, *args, **kwargs):
        return str(self)

    @classmethod
    def all(cls):
        return [target() for target in cls]


class Arg(EnumBase):
    Target = 'use stable (default) dependencies or try latest'
    Pkg = 'create package requirements file'
    Dev = 'create development requirements file'
    Dist = 'create distribution requirements file'

    def help(self):
        return self.value


class Target(EnumBase):
    Stable = auto()
    Latest = auto()


REQUIREMENTS = {
    Arg.Pkg(): requirementsPkg,
    Arg.Dev(): requirementsDev,
    Arg.Dist(): requirementsDist,
}


def generate(*args: Any) -> None:
    parser = argparse.ArgumentParser()

    parser.add_argument(Arg.Target(), help=Arg.Target.help(), choices=Target.all(), nargs='?', default=Target.Stable())
    parser.add_argument(f'--{Arg.Pkg()}', help=Arg.Dev.help(), action="store_true")
    parser.add_argument(f'--{Arg.Dev()}', help=Arg.Dev.help(), action="store_true")
    parser.add_argument(f'--{Arg.Dist()}', help=Arg.Dist.help(), action="store_true")

    parsedArgs = parser.parse_args(args)

    parsedArgsDict = vars(parsedArgs)
    genRegRequests = [parsedArgsDict[reqName] for reqName in REQUIREMENTS.keys()]
    if not any(genRegRequests):
        print("<!> No requirements files were selected!\n(Will fall back to generate package requirements file.)")
        parsedArgsDict[Arg.Pkg()] = True

    target = parsedArgsDict[Arg.Target()]
    separator = '==' if target == Target.Stable() else '>='

    print(f" -> Dependencies versions: {target}")

    def packageToLine(package: Package) -> str:
        return f'{package.name}' \
            + (f' {separator} {package.version}' if package.version else '') \
            + (f'; os_name == "{package.osName}"' if package.osName else '') \
            + '\n'

    def packgesToLines(packages: Packages) -> List[str]:
        return [packageToLine(package) for package in packages.values()]

    def linesToFile(filePath: Path, lines: List[str]) -> None:
        print(f" -> Generating file: '{filePath.name}'")
        with open(filePath, 'w') as fOutput:
            fOutput.writelines(lines)

    for reqName, requirements in REQUIREMENTS.items():
        if not parsedArgsDict[reqName]:
            continue

        linesToFile(requirements.filePath, packgesToLines(requirements.packages))


if __name__ == "__main__":
    generate(*sys.argv[1:])
